--[[

SpyGnome, World of Warcraft addon to gather addon version information from your party players.
Copyright (C) 2007 Rabbit

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

]]

local comm = AceLibrary("AceComm-2.0")
local tablet = AceLibrary("Tablet-2.0")
local L = AceLibrary("AceLocale-2.2"):new("SpyGnome")
local BC = AceLibrary("Babble-Class-2.2")

local _G = getfenv(0)

local addons
local responseTable
local playerNames
local versionStrings = {}

local options = {
	type = "group",
	args = {
		addons = {
			type = "group",
			name = L["Addons"],
			desc = L["What addons to query for."],
			args = {}
		},
	},
}

SpyGnome = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
local SpyGnome = SpyGnome

SpyGnome:RegisterDB("SpyGnomeDB")
SpyGnome:RegisterDefaults("profile", {
	addons = {
		SpyGnome = true,
	},
})
SpyGnome:RegisterChatCommand({"/spygnome"}, options, "SPYGNOME")

SpyGnome.revision = tonumber(("$Revision: 26881 $"):sub(12, -3))
SpyGnome.hasNoColor = true
SpyGnome.hasIcon = "Interface\\Icons\\INV_Misc_Spyglass_01"

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		else
			return {}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				t[k] = nil
			end
			cache[t] = true
		end
	end
end

-- Copied from AceComm who copied from BugGrabber.
function SpyGnome:GetAddonVersionString(addon)
	if not addon then return nil end

	local version = ""
	if AceLibrary:HasInstance(addon) then
		local revision
		version, revision = AceLibrary(addon):GetLibraryVersion()
		version = version .. "-" .. revision
	else
		local revision
		local _G_addon = _G[addon]
		if not _G_addon then
			_G_addon = _G[addon:match("^[^_]+_(.*)$")]
		end
		if _G_addon then
			if rawget(_G_addon, "version") then version = _G_addon.version
			elseif rawget(_G_addon, "Version") then version = _G_addon.Version
			elseif rawget(_G_addon, "VERSION") then version = _G_addon.VERSION
			end
			if type(version) == "function" then version = tostring(select(2, pcall(version()))) end
			local revision = nil
			if rawget(_G_addon, "revision") then revision = _G_addon.revision
			elseif rawget(_G_addon, "Revision") then revision = _G_addon.Revision
			elseif rawget(_G_addon, "REVISION") then revision = _G_addon.REVISION
			elseif rawget(_G_addon, "rev") then revision = _G_addon.rev
			elseif rawget(_G_addon, "Rev") then revision = _G_addon.Rev
			elseif rawget(_G_addon, "REV") then revision = _G_addon.REV
			end
			if type(revision) == "function" then revision = tostring(select(2, pcall(revision()))) end

			if version then version = tostring(version) end
			if revision then revision = tostring(revision) end
			if type(revision) == "string" and type(version) == "string" and version:len() > 0 and not version:find(revision) then
				version = version .. "." .. revision
			end

			if not version and revision then version = revision end
		end

		if _G[addon:upper().."_VERSION"] then
			version = _G[addon:upper() .. "_VERSION"]
		end
		if _G[addon:upper().."_REVISION"] or _G[addon:upper().."_REV"] then
			local revision = _G[addon:upper() .. "_REVISION"] or _G[addon:upper().."_REV"]
			if type(revision) == "string" and type(version) == "string" and version:len() > 0 and not version:find(revision) then
				version = version .. "." .. revision
			end
			if (not version or version == "") and revision then version = revision end
		end

		if not version or version == "" then
			version = GetAddOnMetadata(addon, "Version")
		end
		if not version or version == "" then
			version = IsAddOnLoaded(addon) and true or false
		end
	end
	return version
end

local function VersionCompare(a, b)
	if type(a) == "string" then a = a:trim() end
	if type(b) == "string" then b = b:trim() end
	if a == b then
		return 0
	elseif tonumber(a) and tonumber(b) then
		if tonumber(a) == tonumber(b) then
			return 0
		else
			return tonumber(b) > tonumber(a) and 1 or 2
		end
	else
		a = tostring(a):trim()
		b = tostring(b):trim()
		local numA = a:gsub("%D", "")
		local numB = b:gsub("%D", "")
		numA = tonumber(numA)
		numB = tonumber(numB)
		if type(numA) == "number" and type(numB) == "number" then
			if numA == numB then
				return 0
			else
				return numB > numA and 1 or 2
			end
		else
			if numB and not numA then
				return 1
			else
				return 2
			end
		end
	end
end

function SpyGnome:GetVersionString(addon, otherVersion)
	if not versionStrings[addon] then
		local myVersion = self:GetAddonVersionString(addon)
		local res = VersionCompare(otherVersion, myVersion)
		if res == 0 then -- Same version
			versionStrings[addon] = "|cff00ff00" -- Green
		elseif res == 1 then -- Mine is more recent
			versionStrings[addon] = "|cff0000ff" -- Blue
		elseif res == 2 then -- Other is more recent
			versionStrings[addon] = "|cffff0000" -- Red
		end
	end
	return versionStrings[addon]
end

function SpyGnome:OnInitialize()
	self.version = (self.version or "1") .. "." .. (self.revision or "1")

	local numAddons = GetNumAddOns()
	local i
	for i = 1, numAddons do
		local name = GetAddOnInfo(i)

		options.args.addons.args[name] = {
			type = "toggle",
			name = name,
			desc = string.format(L["Toggle whether to query for %s."], name),
			get = function() return SpyGnome.db.profile.addons[name] end,
			set = function(v)
				SpyGnome.db.profile.addons[name] = v
				SpyGnome:UpdateAddonList()
			end,
		}
	end

	self.OnMenuRequest = options

	self:UpdateAddonList(true)
end

function SpyGnome:UpdateAddonList(dontQuery)
	del(addons)
	addons = new()

	local db = self.db.profile.addons
	local k, v
	for k, v in pairs(db) do
		if db[k] and db[k] == true then
			table.insert(addons, k)
		end
	end

	-- Update in 10 seconds, so we get time to tick/untick a few more addons.
	if not dontQuery then
		self:ScheduleEvent(self.QueryVersions, 10, self)
	end
end

function SpyGnome:OnEnable()
	comm:RegisterAddonVersionReceptor(SpyGnome, "AddonPong")
	self:RegisterBucketEvent({"PARTY_MEMBERS_CHANGED", "RAID_ROSTER_UPDATE"}, 5, "UpdateData")
end

function SpyGnome:OnDataUpdate()
	if not addons or #addons > 4 then return end

	-- Query for new version revisions
	del(responseTable)
	local i, addon
	for i, addon in ipairs(addons) do
		comm:QueryAddonVersion(addon, "GROUP")
	end

	-- Reconstruct the player name table
	del(playerNames)
	playerNames = new()
	local inRaid = UnitInRaid("player")
	local num = inRaid and GetNumRaidMembers() or GetNumPartyMembers()
	for i = 1, num do
		local id = inRaid and "raid"..i or "party"..i
		local name = UnitName(id)
		local class = select(2, UnitClass(id))
		playerNames[name] = "|cff" .. BC:GetHexColor(class) .. name .. "|r"
	end
end

function SpyGnome:AddonPong(player, addon, version)
	if not responseTable then
		responseTable = new()
	end
	if type(responseTable[player]) ~= "table" then
		responseTable[player] = new()
	end
	responseTable[player][addon] = version

	self:UpdateTooltip()
end

function SpyGnome:OnTooltipUpdate()
	if not addons then return end
	if #addons > 4 then
		tablet:SetHint(L["Even though SpyGnomes are very clever and cunning, they can't gather as much information as you have requested. Please keep the number of addon subjects below 5."])
	elseif responseTable then
		local i, v, player, playerAddons

		local tmp = new()
		table.insert(tmp, "text")
		table.insert(tmp, "Player")

		local columns = new()
		for i, v in ipairs(addons) do
			table.insert(tmp, "text"..tostring(i + 1))
			table.insert(tmp, v)
			columns[v] = i + 1
		end
		local cat = tablet:AddCategory("columns", #addons + 1, unpack(tmp))
		del(tmp)
		for player, playerAddons in pairs(responseTable) do
			if type(playerAddons) == "table" then
				tmp = new()
				table.insert(tmp, "text")
				table.insert(tmp, playerNames[player] or player)
				local addon, version
				for addon, version in pairs(playerAddons) do
					if addon and columns[addon] ~= nil then
						table.insert(tmp, "text"..columns[addon])
						table.insert(tmp, self:GetVersionString(addon, tostring(version))..tostring(version).."|r")
					end
				end
				cat:AddLine(unpack(tmp))
				del(tmp)
			end
		end
		del(columns)
		tablet:SetHint(L["|cffff0000Red|r is newer than yours, |cff0000ffBlue|r is older than yours, and |cff00ff00Green|r is the same version."])
	else
		tablet:SetHint(L["The SpyGnome is lurking just behind you, in the shadows. He has no new information for you at this time."])
	end
end

